#ifndef __application_h__
#define __application_h__

#include <string.h>
#include <stdlib.h>

#include "glut_import.h"
#include "list.h"
#include "handled.h"

class Menu;
class Window;

class Application{
  friend class Menu;
  friend class Window;
  protected:
    static List<Window *> HandleQueue;
    static List<Window *> Windows;
    static List<Menu *> Menus;
    static int Running;

    static void updateCallBack();
  public:
    static void run(int argc,char *argv[]);
    static void run();

    static int getRunning(){return Running;};

    static Menu *currentMenu(){
      try{
        int i=glutGetMenu()-1;
        if (i>=0 && i<Menus.getCount())
          return Menus.getItem(i);
        else
          return NULL;
      }
      catch(...){
        return NULL;
      }
    };

    static Window *currentWindow(){
      try{
        int i=glutGetWindow()-1;
        if (i>=0 && i<Windows.getCount())
          return Windows.getItem(i);
        else
          return NULL;
      }
      catch(...){
        return NULL;
      }
    };
};

List<Window *> Application::HandleQueue;
List<Window *> Application::Windows;
List<Menu *> Application::Menus;
int Application::Running=0;

#ifdef STAGE
#  undef STAGE
#endif
#define STAGE 2
#include "window_import.h"
FLUSHCALLBACK

#include "window.h"

#ifdef STAGE
#  undef STAGE
#endif
#define STAGE 3
#include "window_import.h"
FLUSHCALLBACK

void IdleFunc(void){
  Window *w=Application::currentWindow();
  if (w!=NULL && (w->getStyle().CallBacks & WC_IDLE) && w->getVisible())
    w->doIdle();
}

void MenuStatusFunc(int status,int x,int y){
  Window *w=Application::currentWindow();
  if (w!=NULL)
    w->doMenuStatus(status,x,y);
};

void Application::updateCallBack(){
  int i,t=1;
  if (Running){
    for (i=HandleQueue.getCount()-1;i>=0;i--){
      Window *w=HandleQueue[i];
      HandleQueue._delete(i);
      w->handleNeeded();
    }
    for (i=Windows.getCount()-1;i>=0;i--)
      t&=(!Windows[i]->Covered) && (Windows[i]->getStyle().CallBacks & WC_IDLE);
    if (t)
      glutIdleFunc(IdleFunc);
    else
      glutIdleFunc(NULL);
  }
};

void Application::run(int argc,char *argv[]){
  Running=1;
  updateCallBack();
  glutInit(&argc,argv);
  glutMainLoop();
};

void Application::run(){
  Running=1;
  updateCallBack();
  glutMainLoop();
};

#endif
