#ifndef __glarray_h__
#define __glarray_h__

#include <stdio.h>

#include "vect.h"
#include "glut_import.h"

typedef struct {Vect2<float> V;}                                              V2F            ;
typedef struct {Vect3<float> V;}                                              V3F            ;
typedef struct {Vect4<unsigned char> C;Vect2<float> V;}                       C4UB_V2F       ;
typedef struct {Vect4<unsigned char> C;Vect3<float> V;}                       C4UB_V3F       ;
typedef struct {Vect3<float> C;Vect3<float> V;}                               C3F_V3F        ;
typedef struct {Vect3<float> N;Vect3<float> V;}                               N3F_V3F        ;
typedef struct {Vect4<float> C;Vect3<float> N;Vect3<float> V;}                C4F_N3F_V3F    ;
typedef struct {Vect2<float> T;Vect3<float> V;}                               T2F_V3F        ;
typedef struct {Vect4<float> T;Vect4<float> V;}                               T4F_V4F        ;
typedef struct {Vect2<float> T;Vect4<unsigned char> C;Vect3<float> V;}        T2F_C4UB_V3F   ;
typedef struct {Vect2<float> T;Vect3<float> C;Vect3<float> V;}                T2F_C3F_V3F    ;
typedef struct {Vect2<float> T;Vect3<float> N;Vect3<float> V;}                T2F_N3F_V3F    ;
typedef struct {Vect2<float> T;Vect4<float> C;Vect3<float> N;Vect3<float> V;} T2F_C4F_N3F_V3F;
typedef struct {Vect4<float> T;Vect4<float> C;Vect3<float> N;Vect3<float> V;} T4F_C4F_N3F_V4F;

template<GLenum format,class elem> class InterleavedArray: public List<elem>{
  public:
    void bind(){
      glInterleavedArrays(format,0,BaseList<elem>::Data);
    };

    void drawElement(int index){
      CHECKLISTRANGE(index,BaseList<elem>::Count);
      bind();
      glArrayElement(index);
    };

    void drawElements(List<unsigned char> *list,int first,int count,GLenum mode){
      glDrawElements(mode,count,GL_UNSIGNED_BYTE,&list[first]);
    };

    void drawElements(List<unsigned short> *list,int first,int count,GLenum mode){
      glDrawElements(mode,count,GL_UNSIGNED_SHORT,&list[first]);
    };

    void drawElements(List<unsigned int> *list,int first,int count,GLenum mode){
      glDrawElements(mode,count,GL_UNSIGNED_INT,&list[first]);
    };

    template<class indexelem> void drawElements(List<indexelem> *list,GLenum mode){
      bind();
      drawElements(list,0,list.getCount(),mode);
    };

    void drawElements(int first,int count,GLenum mode){
      bind();
      glDrawArrays(mode,first,count);
    };

    void drawElements(GLenum mode){
      drawElements(0,BaseList<elem>::Count,mode);
    };
};

class InterleavedArray_V3F            : public InterleavedArray<GL_V3F            ,V3F            >{};
class InterleavedArray_C4UB_V2F       : public InterleavedArray<GL_C4UB_V2F       ,C4UB_V2F       >{};
class InterleavedArray_C4UB_V3F       : public InterleavedArray<GL_C4UB_V3F       ,C4UB_V3F       >{};
class InterleavedArray_C3F_V3F        : public InterleavedArray<GL_C3F_V3F        ,C3F_V3F        >{};
class InterleavedArray_N3F_V3F        : public InterleavedArray<GL_N3F_V3F        ,N3F_V3F        >{};
class InterleavedArray_C4F_N3F_V3F    : public InterleavedArray<GL_C4F_N3F_V3F    ,C4F_N3F_V3F    >{};
class InterleavedArray_T2F_V3F        : public InterleavedArray<GL_T2F_V3F        ,T2F_V3F        >{};
class InterleavedArray_T4F_V4F        : public InterleavedArray<GL_T4F_V4F        ,T4F_V4F        >{};
class InterleavedArray_T2F_C4UB_V3F   : public InterleavedArray<GL_T2F_C4UB_V3F   ,T2F_C4UB_V3F   >{};
class InterleavedArray_T2F_C3F_V3F    : public InterleavedArray<GL_T2F_C3F_V3F    ,T2F_C3F_V3F    >{};
class InterleavedArray_T2F_N3F_V3F    : public InterleavedArray<GL_T2F_N3F_V3F    ,T2F_N3F_V3F    >{};
class InterleavedArray_T2F_C4F_N3F_V3F: public InterleavedArray<GL_T2F_C4F_N3F_V3F,T2F_C4F_N3F_V3F>{};
class InterleavedArray_T4F_C4F_N3F_V4F: public InterleavedArray<GL_T4F_C4F_N3F_V4F,T4F_C4F_N3F_V4F>{};

#endif

