#ifndef __list_h__
#define __list_h__

#include <stdlib.h>
#include <stdio.h>

#define LASSERT(_c,_s) if (!(_c)) { \
  printf("Error: %s\n",_s);         \
  exit(0xFFFFFFFF);                 \
}

#define CHECKLISTRANGE(_index,_count) if((_index)<0 || (_index)>=(_count)){ \
  printf("Error: list index out of bound (%i)\n",_index);                   \
  exit(0xFFFFFFFF);                                                         \
}

template<class elem> class BaseList{
  protected:
    elem *Data;
    int Capacity,Count;
  public:
    BaseList(){
      Data=NULL;
      Capacity=0;
      Count=0;
    };

    int getCapacity(){
      return Capacity;
    };

    void setCapacity(int n){
      if (n<Count)
        n=Count;
      if (Capacity<n || n<Capacity/2){
        Capacity=3*n/2;
        Data=(elem *) realloc(Data,Capacity*sizeof(elem));
        LASSERT(Capacity==0 || Data!=NULL,"list memory allocation failed");
      }
    };

    int getCount(){
      return Count;
    };

    elem getItem(int index){
      CHECKLISTRANGE(index,Count);
      return Data[index];
    };

    elem operator[](int index){
      CHECKLISTRANGE(index,Count);
      return Data[index];
    };

    int contains(elem x){
      return indexOf(x)==-1?0:1;
    };

    int indexOf(elem x){
      int i;
      for (i=0;i<Count;i++)
        if (Data[i]==x)
          return i;
      return -1;
    };
};

template<class elem> class List: public BaseList<elem>{
  public:
    virtual ~List(){
      if (BaseList<elem>::Data!=NULL)
        free(BaseList<elem>::Data);
    };

    elem *getData(){
      return BaseList<elem>::Data;
    };

    void setItem(elem x,int index){
      CHECKLISTRANGE(index,BaseList<elem>::Count);
      BaseList<elem>::Data[index]=x;
    };

    elem &operator[](int index){
      CHECKLISTRANGE(index,BaseList<elem>::Count);
      return BaseList<elem>::Data[index];
    };

    void insert(elem x,int index){
      int i;
      CHECKLISTRANGE(index,BaseList<elem>::Count+1);
      BaseList<elem>::setCapacity(index+1);
      for (i=BaseList<elem>::Count-1;i>=index;i--)
        BaseList<elem>::Data[i+1]=BaseList<elem>::Data[i];
      BaseList<elem>::Data[index]=x;
      BaseList<elem>::Count++;
    };

    int add(elem x){
      insert(x,BaseList<elem>::Count);
      return BaseList<elem>::Count-1;
    };

    void _delete(int index){
      int i;
      CHECKLISTRANGE(index,BaseList<elem>::Count);
      for (i=index;i<BaseList<elem>::Count-1;i++)
        BaseList<elem>::Data[i]=BaseList<elem>::Data[i+1];
      BaseList<elem>::Count--;
      setCapacity(BaseList<elem>::Count);
    };

    void exchange(int index1,int index2){
      elem x;
      CHECKLISTRANGE(index1,BaseList<elem>::Count);
      CHECKLISTRANGE(index2,BaseList<elem>::Count);
      x=BaseList<elem>::Data[index1];
      BaseList<elem>::Data[index1]=BaseList<elem>::Data[index2];
      BaseList<elem>::Data[index2]=x;
    };

    void remove(elem x){
      int i;
      for (i=BaseList<elem>::Count-1;i>=0;i--)
        if (BaseList<elem>::Data[i]==x)
          _delete(i);
    };

    void clear(){
      if (BaseList<elem>::Count>0){
        BaseList<elem>::Count=0;
        BaseList<elem>::setCapacity(0);
      }
    };
};

template<class elem> class ObjectList: public BaseList<elem>{
  void markModified(){
    if (!Modified){
      changing();
      Modified=1;
    }
  };
  protected:
    int UpdateCount,Modified;

    virtual void changed(){};
    virtual void changing(){};
  public:
    ObjectList():BaseList<elem>(){
      UpdateCount=0;
      Modified=0;
    };

    virtual ~ObjectList(){
      clear();
    };

    virtual void beginUpdate(){
      UpdateCount++;
    };

    virtual void endUpdate(){
      if (UpdateCount<=0){
        printf("Negative updates count");
        throw "Negative updates count";
      }
      UpdateCount--;
      if (Modified && !UpdateCount){
        changed();
        Modified=0;
      }
    };

    virtual int add(elem x){
      insert(x,BaseList<elem>::Count);
      return BaseList<elem>::Count-1;
    };

    virtual void setItem(elem x,int index){
      CHECKLISTRANGE(index,BaseList<elem>::Count);
      if (BaseList<elem>::Data[index]!=x){
        beginUpdate();
        try{
          markModified();
          BaseList<elem>::Data[index]=x;
        }
        catch(...){
          endUpdate();
          throw;
        }
        endUpdate();
      }
    };

    virtual void insert(elem x,int index){
      int i;
      CHECKLISTRANGE(index,BaseList<elem>::Count+1);
      beginUpdate();
      try{
        markModified();
        BaseList<elem>::setCapacity(index+1);
        for (i=BaseList<elem>::Count-1;i>=index;i--)
          BaseList<elem>::Data[i+1]=BaseList<elem>::Data[i];
        BaseList<elem>::Data[index]=x;
        BaseList<elem>::Count++;
      }
      catch(...){
        endUpdate();
        throw;
      }
      endUpdate();
    };

    virtual void _delete(int index){
      int i;
      CHECKLISTRANGE(index,BaseList<elem>::Count);
      beginUpdate();
      try{
        markModified();
        for (i=index;i<BaseList<elem>::Count-1;i++)
          BaseList<elem>::Data[i]=BaseList<elem>::Data[i+1];
        BaseList<elem>::Count--;
        setCapacity(BaseList<elem>::Count);
      }
      catch(...){
        endUpdate();
        throw;
      }
      endUpdate();
    };

    virtual void exchange(int index1,int index2){
      elem x;
      CHECKLISTRANGE(index1,BaseList<elem>::Count);
      CHECKLISTRANGE(index2,BaseList<elem>::Count);
      x=BaseList<elem>::Data[index1];
      beginUpdate();
      try{
        markModified();
        BaseList<elem>::Data[index1]=BaseList<elem>::Data[index2];
        BaseList<elem>::Data[index2]=x;
      }
      catch(...){
        endUpdate();
        throw;
      }
      endUpdate();
    };

    virtual void remove(elem x){
      int i;
      beginUpdate();
      try{
        for (i=BaseList<elem>::Count-1;i>=0;i--)
          if (BaseList<elem>::Data[i]==x)
            _delete(i);
      }
      catch(...){
        endUpdate();
        throw;
      }
      endUpdate();
    };

    virtual void clear(){
      int i;
      if (BaseList<elem>::Count>0){
        beginUpdate();
        try{
          for (i=BaseList<elem>::Count-1;i>=0;i--)
            _delete(i);
        }
        catch(...){
          endUpdate();
          throw;
        }
        endUpdate();
      }
    };
};

#endif

