#ifndef __matrix_h__
#define __matrix_h__

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#define CHECKMATRIXRANGE(_x,_y,_width,_height) \
if ((_x)<0 || (_y)<0 || (_x)>=(_width) || (_y)>=(_height)){\
  printf("Error: matrix index out of bound (%i,%i)\n",_x,_y);\
  exit(0xFFFFFFFF);\
}

template<class elem> class Matrix{
  void alloc(int width,int height){
    assert(height>0 && width>0);
    elem *p=(elem *) malloc(width*height*sizeof(elem));
    assert(p!=NULL);
    /*int i,j,m=width<Width?width:Width,n=height<Height?height:Height;
    for (i=0;i<m;i++)
      for (j=0;j<n;j++)
        p[i+j*width]=Data[i+j*Width];*/
    free(Data);
    Data=p;
    Width=width;
    Height=height;
  };

  protected:
    elem *Data;
    int Width,Height;

  public:
    Matrix(int width,int height){
      Data=NULL;
      alloc(width,height);
    };

    Matrix(Matrix &m){
      int i;
      Data=NULL;
      alloc(m.Width,m.Height);
      for (i=Width*Height-1;i>=0;i--)
        Data[i]=m.Data[i];
    };

    Matrix(int width,int height,elem value){
      int i;
      Data=NULL;
      alloc(width,height);
      for (i=Width*Height-1;i>=0;i--)
        Data[i]=value;
    };

    virtual ~Matrix(){
      if (Data!=NULL)
        free(Data);
    };

    int getWidth(){return Width;};
    void setWidth(int width){
      alloc(width,Height);
    };

    int getHeight(){return Height;};
    void setHeight(int height){
      alloc(Width,height);
    };

    elem getItem(int x,int y){
      CHECKMATRIXRANGE(x,y,Width,Height);
      return Data[x+y*Width];
    };

    void setItem(int x,int y,elem value){
      CHECKMATRIXRANGE(x,y,Width,Height);
      Data[x+y*Width]=value;
    };

    elem operator()(int x,int y) const{
      CHECKMATRIXRANGE(x,y,Width,Height);
      return Data[x+y*Width];
    };

    elem &operator()(int x,int y){
      CHECKMATRIXRANGE(x,y,Width,Height);
      return Data[x+y*Width];
    };

    Matrix &operator=(Matrix &m){
      int i;
      alloc(m.Width,m.Height);
      for (i=Width*Height-1;i>=0;i--)
        Data[i]=m.Data[i];
      return *this;
    };
};

#endif
