#ifndef __maze_h__
#define __maze_h__

#include <limits.h>
#include <math.h>
#include <time.h>

#include "matrix.h"

typedef enum{NORTH,WEST,SOUTH,EAST} Direction;

Direction turnRight(Direction direction){
  switch (direction){
    case NORTH:return WEST;
    case WEST:return SOUTH;
    case SOUTH:return EAST;
    case EAST:return NORTH;
    default:return NORTH;
  };
};

Direction turnLeft(Direction direction){
  switch (direction){
    case NORTH:return EAST;
    case WEST:return NORTH;
    case SOUTH:return WEST;
    case EAST:return SOUTH;
    default:return NORTH;
  };
};

Direction backWard(Direction direction){
  switch (direction){
    case NORTH:return SOUTH;
    case WEST:return EAST;
    case SOUTH:return NORTH;
    case EAST:return WEST;
    default:return NORTH;
  };
};

struct MazeCell{
  int North,West;
};

const MazeCell EmptyCell={0,0};
const MazeCell FullCell={1,1};

class MazeGrid: public Matrix<MazeCell>{
  public:
    MazeGrid(int width,int height):Matrix<MazeCell>(width,height,EmptyCell){};

    int canMove(int x,int y,Direction direction,int back=0){
      if (back)
        direction=backWard(direction);
      switch(direction){
        case NORTH:
          return y>0 && !getItem(x,y).North;
        case WEST:
          return x>0 && !getItem(x,y).West;
        case SOUTH:
          return y<Height-1 && !getItem(x,y+1).North;
        case EAST:
          return x<Width-1 && !getItem(x+1,y).West;
        default:
          return 0;
      }
    };

    int dist(int x1,int y1,int x2,int y2){
      Matrix<int> M(Width,Height,INT_MAX);
      int i,j,k,l;
      M(x1,y1)=0;
      do{
        k=0;
        for (i=0;i<Width;i++)
          for (j=0;j<Height;j++){
            l=M(i,j);
            if (canMove(i,j,NORTH) && l>M(i,j-1))
              l=M(i,j-1)+1;
            if (canMove(i,j,WEST) && l>M(i-1,j))
              l=M(i-1,j)+1;
            if (canMove(i,j,SOUTH) && l>M(i,j+1))
              l=M(i,j+1)+1;
            if (canMove(i,j,EAST) && l>M(i+1,j))
              l=M(i+1,j)+1;
            if (M(i,j)>l){
              k=1;
              M(i,j)=l;
            }
          }
      }while (k && M(x2,y2)==INT_MAX);
      return (M(x2,y2)==INT_MAX)?-1:M(x2,y2);
    };

    void randAdd(){
      MazeCell *C=Data;
      int l,u=0,v;
      for (l=0;l<Width*Height;l++){
        if (!C->North) u++;
        if (!C->West) u++;
        C++;
      }
      v=1+(rand() % u);C=Data;u=0;
      while(1){
        if (!C->North) u++;
        if (u==v){C->North=1;return;}
        if (!C->West) u++;
        if (u==v){C->West=1;return;}
        C++;
      }
    };

    void randRemove(){
      MazeCell *C=Data;
      int l,u=0,v;
      for (l=0;l<Width*Height;l++){
        if (C->North) u++;
        if (C->West) u++;
        C++;
      }
      v=1+(rand() % u);C=Data;u=0;
      while(1){
        if (C->North) u++;
        if (u==v){C->North=0;return;}
        if (C->West) u++;
        if (u==v){C->West=0;return;}
        C++;
      }
    };

    void generate(int max){
      Matrix<MazeCell> Best=*this;
      int k=0,l,m=-1,n,f=0;
      srand(time(NULL));
      printf("Generation du labyrinthe, veuillez patienter... ");
      for (n=0;n<max;n++){
        l=dist(0,0,Width-1,Height-1);
        if (l>=0){
          if (l>m){
            m=l;
            Best=*this;
          }
          randAdd();
          f++;
        }else{
          randRemove();
          f--;
        }
      };
      printf("termine!\n");
      Matrix<MazeCell>::operator =(Best);
    };
};

#endif
