#ifndef __mazemesh_h__
#define __mazemesh_h__

#include "vect.h"
#include "maze.h"
#include "interleavedarray.h"
#include "texture.h"
#include "light.h"

class QuadMesh: public InterleavedArray_T2F_N3F_V3F{
  public:
    Texture_2D *Tex;

    QuadMesh(char *textureBitmapFileName){
      Tex=new Texture_2D;
      Tex->loadBitmapFile(textureBitmapFileName);
    };

    virtual ~QuadMesh(){
      delete Tex;
      clear();
    };

    void addQuad(Vect3<float> M,Vect3<float> U,Vect3<float> V,float u=1,float v=1,float u0=0,float v0=0){
      T2F_N3F_V3F vertex;
      vertex.N=normalize(U^V);
      vertex.T=vect(u0,v0);
      vertex.V=M;
      InterleavedArray_T2F_N3F_V3F::add(vertex);
      vertex.T=vect(u0,v0+v);
      vertex.V=M+V;
      InterleavedArray_T2F_N3F_V3F::add(vertex);
      vertex.T=vect(u0+u,v0+v);
      vertex.V=M+U+V;
      InterleavedArray_T2F_N3F_V3F::add(vertex);
      vertex.T=vect(u0+u,v0);
      vertex.V=M+U;
      InterleavedArray_T2F_N3F_V3F::add(vertex);
    };

    void bind(){
      glEnable(GL_TEXTURE_2D);
      Tex->bind();
      drawElements(GL_QUADS);
      glDisable(GL_TEXTURE_2D);
    }
};

class MazeMesh{
  QuadMesh *Walls,*Floor,*Sky[5];

  public:
    MazeMesh(MazeGrid *grid){
      int i,x,y,z;
#define TexName() "lostvalley"
      Floor=new QuadMesh("floor.bmp");
      Walls=new QuadMesh("wall.bmp");
      Sky[0]=new QuadMesh(TexName() "_up.bmp");
      Sky[1]=new QuadMesh(TexName() "_north.bmp");
      Sky[2]=new QuadMesh(TexName() "_west.bmp");
      Sky[3]=new QuadMesh(TexName() "_south.bmp");
      Sky[4]=new QuadMesh(TexName() "_east.bmp");
      float F=0.1;
      Walls->addQuad(vect(F,0.f,F),
                     vect(grid->getWidth()-2*F,0.f,0.f),
                     vect(0.f,0.7f,0.f),
                     grid->getWidth()-2*F);
      Walls->addQuad(vect(F,0.f,grid->getHeight()-F),
                     vect(grid->getWidth()-2*F,0.f,0.f),
                     vect(0.f,0.7f,0.f),
                     grid->getWidth()-2*F);
      Walls->addQuad(vect(F,0.f,F),
                     vect(0.f,0.f,grid->getHeight()-2*F),
                     vect(0.f,0.7f,0.f),
                     grid->getHeight()-2*F);
      Walls->addQuad(vect(grid->getWidth()-F,0.f,F),
                     vect(0.f,0.f,grid->getHeight()-2*F),
                     vect(0.f,0.7f,0.f),
                     grid->getHeight()-2*F);
      for (x=1;x<grid->getWidth();x++)
        for (y=0;y<grid->getHeight();y++){
          z=y;
          while (y<grid->getHeight() && (*grid)(x,y).West)
            y++;
          if (y>z){
            Walls->addQuad(vect(x+F,0.f,z+F),vect(0.f,0.f,y-z-2*F),vect(0.,0.7,0.),y-z-2*F);
            Walls->addQuad(vect(x-F,0.f,z+F),vect(0.f,0.f,y-z-2*F),vect(0.,0.7,0.),y-z-2*F);

            Walls->addQuad(vect(x-F,0.f,z+F),vect(2*F,0.f,0.f),vect(0.,0.7,0.),2*F);
            Walls->addQuad(vect(x-F,0.f,y-F),vect(2*F,0.f,0.f),vect(0.,0.7,0.),2*F);
          }
        }
      for (y=1;y<grid->getHeight();y++)
        for (x=0;x<grid->getWidth();x++){
          z=x;
          while (x<grid->getWidth() && (*grid)(x,y).North)
            x++;
          if (x>z){
            Walls->addQuad(vect(z+F,0.f,y-F),vect(x-z-2*F,0.f,0.f),vect(0.,0.7,0.),x-z-2*F);
            Walls->addQuad(vect(z+F,0.f,y+F),vect(x-z-2*F,0.f,0.f),vect(0.,0.7,0.),x-z-2*F);

            Walls->addQuad(vect(z+F,0.f,y-F),vect(0.f,0.f,2*F),vect(0.,0.7,0.),2*F,1,-F);
            Walls->addQuad(vect(x-F,0.f,y-F),vect(0.f,0.f,2*F),vect(0.,0.7,0.),2*F,1,-F);
          }
        }
      Floor->addQuad(vect(0,0,0),vect(grid->getWidth(),0,0),vect(0,0,grid->getHeight()),grid->getWidth(),grid->getHeight());
      float D=500;
      for (i=0;i<5;i++){
        Sky[i]->Tex->setWrapS(GL_CLAMP);
        Sky[i]->Tex->setWrapT(GL_CLAMP);
      }
      Sky[0]->addQuad(vect(-D,D,-D),vect(2*D,0.f,0.f),vect(0.f,0.f,2*D));
      Sky[1]->addQuad(vect(D,-D,D),vect(-2*D,0.f,0.f),vect(0.f,2*D,0.f));
      Sky[2]->addQuad(vect(D,-D,-D),vect(0.f,0.f,2*D),vect(0.f,2*D,0.f));
      Sky[3]->addQuad(vect(-D,-D,-D),vect(2*D,0.f,0.f),vect(0.f,2*D,0.f));
      Sky[4]->addQuad(vect(-D,-D,D),vect(0.f,0.f,-2*D),vect(0.f,2*D,0.f));
    };

    virtual ~MazeMesh(){
      int i;
      delete Walls;
      delete Floor;
      for (i=0;i<5;i++)
        delete Sky[i];
    };

    void bind(){
      int i;
      for (i=0;i<5;i++)
        Sky[i]->bind();
      glutReportErrors();
      Floor->bind();
      Walls->bind();
      glDisable(GL_LIGHTING);
    };
};

#endif
