#ifndef __texture_h__
#define __texture_h__

#include <stdio.h>
#include <limits.h>

#include "glut_import.h"

#include "bitmap.h"
#include "handled.h"

class Texture: public Handled<GLuint,INT_MAX>{
  protected:
    GLenum Class;
    unsigned int Format,MinFilter,MagFilter,WrapS;

    virtual GLenum getClass(){return 0;};

    void createHandle(){glGenTextures(1,&Handle);};
    void deleteHandle(){glDeleteTextures(1,&Handle);};

    template<class elem> GLenum bitmapDataType(Bitmap<elem> *){
      printf("Unsupported Bitmap data type\n");
      exit(0xFFFFFFFF);
    };
  public:
    Texture(unsigned int aFormat=GL_RGB,
            unsigned int aMinFilter=GL_LINEAR_MIPMAP_LINEAR,
            unsigned int aMagFilter=GL_LINEAR):Format(aFormat),MinFilter(aMinFilter),MagFilter(aMagFilter){
      Class=0;
    };

    ~Texture(){
      freeHandle();
    };

    virtual void bind(){
      if (Class==0)
        Class=getClass();
      glBindTexture(Class,getHandle());
      glTexParameterf(Class,GL_TEXTURE_MIN_FILTER,MinFilter);
      glTexParameterf(Class,GL_TEXTURE_MAG_FILTER,MagFilter);
    };

    unsigned int getMinFilter(){return MinFilter;};
    void setMinFilter(unsigned int aFilter){MinFilter=aFilter;};
    unsigned int getMagFilter(){return MagFilter;};
    void setMagFilter(unsigned int aFilter){MagFilter=aFilter;};
};

template<> GLenum Texture::bitmapDataType(Bitmap<unsigned char > *){return GL_UNSIGNED_BYTE;};
template<> GLenum Texture::bitmapDataType(Bitmap<signed char   > *){return GL_BYTE;};
template<> GLenum Texture::bitmapDataType(Bitmap<unsigned short> *){return GL_UNSIGNED_SHORT;};
template<> GLenum Texture::bitmapDataType(Bitmap<signed short  > *){return GL_SHORT;};
template<> GLenum Texture::bitmapDataType(Bitmap<unsigned int  > *){return GL_UNSIGNED_INT;};
template<> GLenum Texture::bitmapDataType(Bitmap<signed int    > *){return GL_INT;};
template<> GLenum Texture::bitmapDataType(Bitmap<float         > *){return GL_FLOAT;};

class Texture_1D : public Texture{
  protected:
    unsigned int WrapS;

    virtual GLenum getClass(){return GL_TEXTURE_1D;};
  public:
    Texture_1D(unsigned int aFormat=GL_RGB,
               unsigned int aWrap=GL_REPEAT,
               unsigned int aMinFilter=GL_LINEAR_MIPMAP_LINEAR,
               unsigned int aMagFilter=GL_LINEAR):Texture(aFormat,aMinFilter,aMagFilter),WrapS(aWrap){};

    void bind(){
      Texture::bind();
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,WrapS);
    };

    unsigned int getWrapS(){return WrapS;};
    void setWrapS(unsigned int aWrap){WrapS=aWrap;};
};

class Texture_2D : public Texture_1D{
  template<class elem> void loadBitmap(Bitmap<elem> *,unsigned int);
  protected:
    unsigned int WrapT;

    GLenum getClass(){return GL_TEXTURE_2D;};
  public:
    Texture_2D(unsigned int aFormat=GL_RGB,
               unsigned int aMagFilter=GL_LINEAR,
               unsigned int aWrap=GL_REPEAT,
               unsigned int aMinFilter=GL_LINEAR_MIPMAP_LINEAR):Texture_1D(aFormat,aWrap,aMinFilter,aMagFilter),WrapT(aWrap){};

    void loadBitmapFile(char *filename,int useMipmaps=1){
      Bitmap<unsigned char> b;
      b.loadFile(filename);
      try{
        loadBitmap(&b,useMipmaps);
      }
      catch(char *s){
        printf("Error while loading bitmap image \"%s\":\n%s\n",filename,s);
        exit(0xFFFFFFFF);
      }
    };

    void bind(){
      Texture_1D::bind();
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,WrapT);
    };

    unsigned int getWrapT(){return WrapT;};
    void setWrapT(unsigned int aWrap){WrapT=aWrap;};
};

int isPowerOfTwo(int i){
  for (;i;i>>=1)
    if (i!=1 && (i&1)==1)
      return 0;
  return 1;
}

template<class elem> void Texture_2D::loadBitmap(Bitmap<elem> *p,unsigned int useMipmaps){
  bind();
  glPixelStorei(GL_UNPACK_ALIGNMENT,1);
  if (useMipmaps)
    gluBuild2DMipmaps(GL_TEXTURE_2D,Format,p->getWidth(),p->getHeight(),GL_RGB,bitmapDataType(p),p->getData());
  else{
    if (!isPowerOfTwo(p->getWidth()) || !isPowerOfTwo(p->getHeight()))
      throw "Image size is not a power of 2";
    glTexImage2D(GL_TEXTURE_2D,0,Format,p->getWidth(),p->getHeight(),0,GL_RGB,bitmapDataType(p),p->getData());
  }
};

#endif

