#ifndef __vect_h__
#define __vect_h__

#include <math.h>

#pragma pack(1)
template<class elem=float> struct Vect4{
  elem A,B,C,D;
  template<class elem2> operator Vect4<elem2>(){
    Vect4<elem2> q;
    q.A=A;
    q.B=B;
    q.C=C;
    q.D=D;
    return q;
  };
  template<typename elem2> operator elem2(){
    elem2 q;
    q.A=A;
    q.B=B;
    q.C=C;
    return q;
  };
};
template<class elem=float> struct Vect3{
  elem A,B,C;
  template<class elem2> operator Vect3<elem2>(){
    Vect3<elem2> q;
    q.A=A;
    q.B=B;
    q.C=C;
    return q;
  };
  operator Vect4<unsigned char>(){
    Vect4<unsigned char> q;
    q.A=A;
    q.B=B;
    q.C=C;
    q.D=255;
  };
  template<class elem2> operator Vect4<elem2>(){
    Vect4<elem2> q;
    q.A=A;
    q.B=B;
    q.C=C;
    q.D=1;
    return q;
  };
};
template<class elem=float> struct Vect2{
  elem A,B;
  template<class elem2> operator Vect2<elem2>(){
    Vect2<elem2> q;
    q.A=A;
    q.B=B;
    return q;
  };
  template<class elem2> operator Vect3<elem2>(){
    Vect3<elem2> q;
    q.A=A;
    q.B=B;
    q.C=0;
    return q;
  };
};
#pragma pack()

template<class elem> Vect2<elem> vect(elem a,elem b){
  Vect2<elem> q;
  q.A=a;
  q.B=b;
  return q;
}
template<class elem> Vect3<elem> vect(elem a,elem b,elem c){
  Vect3<elem> q;
  q.A=a;
  q.B=b;
  q.C=c;
  return q;
}
template<class elem> Vect4<elem> vect(elem a,elem b,elem c,elem d){
  Vect4<elem> q;
  q.A=a;
  q.B=b;
  q.C=c;
  q.D=d;
  return q;
}

template<class elem> Vect2<elem> operator +(const Vect2<elem> a,const Vect2<elem> b){
  Vect2<elem> q;
  q.A=a.A+b.A;
  q.B=a.B+b.B;
  return q;
}
template<class elem> Vect3<elem> operator +(const Vect3<elem> a,const Vect3<elem> b){
  Vect3<elem> q;
  q.A=a.A+b.A;
  q.B=a.B+b.B;
  q.C=a.C+b.C;
  return q;
}
template<class elem> Vect4<elem> operator +(const Vect4<elem> a,const Vect4<elem> b){
  Vect4<elem> q;
  q.A=a.A+b.A;
  q.B=a.B+b.B;
  q.C=a.C+b.C;
  q.D=a.D+b.D;
  return q;
}

template<class elem> Vect2<elem> operator -(const Vect2<elem> a,const Vect2<elem> b){
  Vect2<elem> q;
  q.A=a.A-b.A;
  q.B=a.B-b.B;
  return q;
}
template<class elem> Vect3<elem> operator -(const Vect3<elem> a,const Vect3<elem> b){
  Vect3<elem> q;
  q.A=a.A-b.A;
  q.B=a.B-b.B;
  q.C=a.C-b.C;
  return q;
}
template<class elem> Vect4<elem> operator -(const Vect4<elem> a,const Vect4<elem> b){
  Vect4<elem> q;
  q.A=a.A-b.A;
  q.B=a.B-b.B;
  q.C=a.C-b.C;
  q.D=a.D-b.D;
  return q;
}

template<class elem> elem operator *(const Vect2<elem> a,const Vect2<elem> b){
  return a.A*b.A+a.B*b.B;
}
template<class elem> elem operator *(const Vect3<elem> a,const Vect3<elem> b){
  return a.A*b.A+a.B*b.B+a.C*b.C;
}
template<class elem> elem operator *(const Vect4<elem> a,const Vect4<elem> b){
  return a.A*b.A+a.B*b.B+a.C*b.C+a.D*b.D;
}

template<class elem> Vect2<elem> operator *(elem k,const Vect2<elem> a){
  Vect2<elem> q;
  q.A=a.A*k;
  q.B=a.B*k;
  return q;
}
template<class elem> Vect3<elem> operator *(elem k,const Vect3<elem> a){
  Vect3<elem> q;
  q.A=a.A*k;
  q.B=a.B*k;
  q.C=a.C*k;
  return q;
}
template<class elem> Vect4<elem> operator *(elem k,const Vect4<elem> a){
  Vect4<elem> q;
  q.A=a.A*k;
  q.B=a.B*k;
  q.C=a.C*k;
  q.D=a.D*k;
  return q;
}

template<class elem> Vect3<elem> operator ^(const Vect3<elem> a,const Vect3<elem> b){
  Vect3<elem> q;
  q.A=a.B*b.C-a.C*b.B;
  q.B=a.C*b.A-a.A*b.C;
  q.C=a.A*b.B-a.B*b.A;
  return q;
}

template<class elem> elem norm(const Vect2<elem> a){
  return sqrt(a.A*a.A+a.B*a.B);
}
template<class elem> elem norm(const Vect3<elem> a){
  return sqrt(a.A*a.A+a.B*a.B+a.C*a.C);
}
template<class elem> elem norm(const Vect4<elem> a){
  return sqrt(a.A*a.A+a.B*a.B+a.C*a.C+a.D*a.D);
}

template<class elem> Vect2<elem> normalize(Vect2<elem> q){
  float r=1/(0.0000000001+norm(q));
  return r*q;
}
template<class elem> Vect3<elem> normalize(Vect3<elem> q){
  float r=1/(0.0000000001+norm(q));
  return r*q;
}
template<class elem> Vect4<elem> normalize(Vect4<elem> q){
  float r=1/(0.0000000001+norm(q));
  return r*q;
}

#endif
