#ifndef __window_h__
#define __window_h__

#include "menu.h"

#define randf() (rand()/(1+(double)RAND_MAX))

#define WC_STANDARD     0x0000
#define WC_IDLE         0x0001
#define WC_KEYBOARD     0x0002
#define WC_MOUSE        0x0004
#define WC_BUTTONBOX    0x0008
#define WC_DIALS        0x0010
#define WC_SPACEBALL    0x0020
#define WC_TABLET       0x0040

#define WF_NAME         0x0001
#define WF_PARENT       0x0002
#define WF_POSITION     0x0004
#define WF_SIZE         0x0008
#define WF_BOUNDS       (WF_POSITION | WF_SIZE)
#define WF_DISPLAYMODE  0x0010
#define WF_CALLBACKS    0x0020
#define WF_MENUS        0x0040
#define WF_FULLSCREEN   0x0080
#define WF_VISIBLE      0x0100

class Window: public Handled<int,-1>, public ObjectList<Window *>{
  friend void IdleFunc(void);
  friend void MenuStatusFunc(int status,int x,int y);
  friend class Application;

#ifdef STAGE
#  undef STAGE
#endif
#define STAGE 1
#include "window_import.h"
  FLUSHCALLBACK

  virtual void doDisplay(){onDisplay();};
  virtual void doReshape(int width,int height){onReshape(width,height);};
  virtual void doWindowStatus(int state){
    if (state==GLUT_HIDDEN || state==GLUT_FULLY_COVERED){
      Covered=1;
      Application::updateCallBack();
      onHide();
    }else{
      Covered=0;
      Application::updateCallBack();
      onShow();
    }
  };
  virtual void doMenuStatus(int status,int x,int y){
    if (status==GLUT_MENU_IN_USE)
      onMenuEnter(x,y);
    else
      onMenuLeave(x,y);
  };
  virtual void doIdle(){onIdle();};
  virtual void doKeyboard(unsigned char key,int x,int y){onKeyboard(key,x,y);};
  virtual void doSpecial(int key,int x,int y){onSpecial(key,x,y);};
#if (GLUT_API_VERSION >= 4 || GLUT_XLIB_IMPLEMENTATION >= 13)
  virtual void doKeyboardUp(unsigned char key,int x,int y){onKeyboardUp(key,x,y);};
  virtual void doSpecialUp(int key,int x,int y){onSpecialUp(key,x,y);};
#endif
  virtual void doMouse(int button,int state,int x,int y){
    if (state==GLUT_DOWN)
	    onMouseDown(button,x,y);
    else
      onMouseUp(button,x,y);
  };
  virtual void doEntry(int state){
    if (state==GLUT_ENTERED)
      onMouseEnter();
    else
      onMouseLeave();
  };
  virtual void doMotion(int x,int y){onMouseMotion(0,x,y);};
  virtual void doPassiveMotion(int x,int y){onMouseMotion(1,x,y);};
  virtual void doButtonBox(int button,int state){
    if (state==GLUT_DOWN)
      onButtonBoxDown(button);
    else
      onButtonBoxUp(button);
  };
  virtual void doDials(int dial,int value){onDials(dial,value);};
  virtual void doSpaceballButton(int button,int state){
    if (state==GLUT_DOWN)
      onSpaceballButtonDown(button);
    else
      onSpaceballButtonUp(button);
  };
  virtual void doSpaceballMotion(int x,int y,int z){onSpaceballMotion(x,y,z);};
  virtual void doSpaceballRotate(int x,int y,int z){onSpaceballRotate(x,y,z);};

  virtual void doTabletButton(int button,int state,int x,int y){
    if (state==GLUT_DOWN)
      onTabletButtonDown(button,x,y);
    else
      onTabletButtonUp(button,x,y);
  };
  virtual void doTabletMotion(int x,int y){onTabletMotion(x,y);};

  void bind(){glutSetWindow(getHandle());};

  void updateFlags(unsigned int flags){
    int h=0;
    if (handleAllocated()){
      bind();
      if (WF_NAME & flags)
        glutSetWindowTitle(Name);
      if (WF_PARENT & flags)
        h=1;
      if (WF_POSITION & flags)
        glutPositionWindow(Bounds.X,Bounds.Y);
      if (WF_SIZE & flags)
        glutReshapeWindow(Bounds.Width,Bounds.Height);
      if (WF_DISPLAYMODE & flags)
        h=1;
      if (WF_CALLBACKS & flags){
        Application::updateCallBack();
#       ifdef STAGE
#         undef STAGE
#       endif
#       define STAGE 4
#       include "window_import.h"
        FLUSHCALLBACK
      }
      if (WF_MENUS & flags){
        int i;
        const int T[3]={GLUT_LEFT_BUTTON,GLUT_RIGHT_BUTTON,GLUT_MIDDLE_BUTTON};
        for (i=0;i<3;i++)
          if (PopupMenu[i]!=NULL){
            glutSetMenu(PopupMenu[i]->getHandle());
            glutAttachMenu(T[i]);
          }
      }
      if (WF_FULLSCREEN & flags){
        if (Style.FullScreen && Parent==NULL)
          glutFullScreen();
        else
          updateFlags(WF_BOUNDS);
      }
      if (WF_VISIBLE & flags){
        if (Style.Visible)
          glutShowWindow();
        else
          glutHideWindow();
      }
      if (h)
        recreateHandle();
    }
  };

  public:
    struct Rect{
      int X,Y,Width,Height;
    };

    struct WindowStyle{
      unsigned int DisplayMode,CallBacks;
      int Visible,FullScreen;
    };

    static const Rect DefTopWindowBounds;
    static const Rect DefSubWindowBounds;
    static const WindowStyle DefTopWindowStyle;
    static const WindowStyle DefSubWindowStyle;
  protected:
    char *Name;
    Window *Parent;
    Rect Bounds;
    WindowStyle Style;
    Menu *PopupMenu[3];
    int Covered;

    void createHandle(){
      glutInitDisplayMode(Style.DisplayMode);
      if (Parent==NULL){
        glutInitWindowPosition(Bounds.X,Bounds.Y);
        glutInitWindowSize(Bounds.Width,Bounds.Height);
        Handle=glutCreateWindow(Name);
        if (Style.FullScreen)
          glutFullScreen();
      }else
        Handle=glutCreateSubWindow(Parent->getHandle(),Bounds.X,Bounds.Y,Bounds.Width,Bounds.Height);
      if (Handle==Application::Windows.getCount()+1)
        Application::Windows.add(this);
      else
        Application::Windows.setItem(this,Handle-1);
      updateFlags(WF_CALLBACKS | WF_MENUS | WF_VISIBLE);
      onCreate();
    };

    void deleteHandle(){
      bind();
      onDestroy();
      Application::Windows.setItem(NULL,Handle-1);
      glutDestroyWindow(Handle);
    };
  public:
    Window(char *name,WindowStyle style=DefTopWindowStyle,Rect bounds=DefTopWindowBounds):Parent(NULL),Name(NULL),Style(style),Bounds(bounds){
      setName(name);
      PopupMenu[0]=NULL;
      PopupMenu[1]=NULL;
      PopupMenu[2]=NULL;
      Covered=1;
      Application::HandleQueue.add(this);
      Application::updateCallBack();
    };

    Window(Window *parent,WindowStyle style=DefTopWindowStyle,Rect bounds=DefSubWindowBounds):Parent(parent),Name(NULL),Style(style),Bounds(bounds){
      setName("");
      PopupMenu[0]=NULL;
      PopupMenu[1]=NULL;
      PopupMenu[2]=NULL;
      if (Parent!=NULL)
        Parent->add(this);
      Covered=1;
      Application::HandleQueue.add(this);
      Application::updateCallBack();
    };

    virtual ~Window(){
      if (Parent!=NULL)
        Parent->remove(this);
      if (Name!=NULL)
        free(Name);
      freeHandle();
    };

    char *getName(){return Name;};
    void setName(char *name){
      if (name==NULL)
        name="";
      Name=(char *) malloc(strlen(name)+1);
      strcpy(Name,name);
      updateFlags(WF_NAME);
    };

    Window *getParent(){return Parent;};
    void setParent(Window *parent){
      if (Parent!=parent){
        if (Parent!=NULL)
          Parent->remove(this);
        Parent=parent;
        if (Parent!=NULL)
          Parent->add(this);
      }
      recreateHandle();
    };

    int getX(){
      bind();
      return glutGet(GLUT_WINDOW_X);
    };
    void setX(const int x){
      Bounds.X=x;
      updateFlags(WF_POSITION);
    };

    int getY(){
      bind();
      return glutGet(GLUT_WINDOW_Y);
    };
    void setY(const int y){
      Bounds.Y=y;
      updateFlags(WF_POSITION);
    };

    int getWidth(){
      bind();
      return glutGet(GLUT_WINDOW_WIDTH);
    };
    void setWidth(const int width){
      Bounds.Width=width;
      Bounds.Height=getHeight();
      updateFlags(WF_SIZE);
    };

    int getHeight(){
      bind();
      return glutGet(GLUT_WINDOW_HEIGHT);
    };
    void setHeight(const int height){
      Bounds.Width=getWidth();
      Bounds.Height=height;
      updateFlags(WF_SIZE);
    };

    void setPosition(int x,int y){
      Bounds.X=x;
      Bounds.Y=y;
      updateFlags(WF_POSITION);
    };

    void setSize(int width,int height){
      Bounds.Width=width;
      Bounds.Height=height;
      updateFlags(WF_SIZE);
    };

    Rect getBounds(){
      Rect R;
      R.X=getX();
      R.Y=getY();
      R.Width=getWidth();
      R.Height=getHeight();
      return R;
    };
    void setBounds(Rect bounds){
      Bounds=bounds;
      updateFlags(WF_BOUNDS);
    };
    void setBounds(int x,int y,int width,int height){
      Bounds.X=x;
      Bounds.Y=y;
      Bounds.Width=width;
      Bounds.Height=height;
      updateFlags(WF_BOUNDS);
    };

    unsigned int getDisplayMode(){return Style.DisplayMode;};
    void setDisplayMode(unsigned int displayMode){
      if (Style.DisplayMode!=displayMode){
        Style.DisplayMode=displayMode;
        recreateHandle();
      }
    };

    unsigned int getCallBacks(){return Style.CallBacks;};
    void setCallBacks(unsigned int callBacks){
      if (Style.CallBacks!=callBacks){
        Style.CallBacks=callBacks;
        updateFlags(WF_CALLBACKS);
      }
    };
    void addCallBacks(unsigned int callBacks){
      Style.CallBacks=Style.CallBacks | callBacks;
      updateFlags(WF_CALLBACKS);
    };
    void removeCallBacks(unsigned int callBacks){
      Style.CallBacks=Style.CallBacks & ~callBacks;
      updateFlags(WF_CALLBACKS);
    };

    int getFullScreen(){return Style.FullScreen;};
    void setFullScreen(int fullScreen){
      if (Style.FullScreen!=fullScreen){
        Style.FullScreen=fullScreen;
        updateFlags(WF_FULLSCREEN);
      }
    };

    WindowStyle getStyle(){return Style;};
    void setStyle(WindowStyle style){
      if (Style.DisplayMode!=style.DisplayMode){
        Style=style;
        recreateHandle();
        setCallBacks(style.CallBacks);
      }else{
        setFullScreen(style.FullScreen);
        setCallBacks(style.CallBacks);
      }
    };

    Menu *getPopupMenu(int button=1){
      return PopupMenu[button];
    };

    void setPopupMenu(Menu *menu,int button=1){
      if (PopupMenu[button]!=menu){
        if (PopupMenu[button]!=NULL)
          freeHandle();
        PopupMenu[button]=menu;
        updateFlags(WF_MENUS);
      }
    };

    int getVisible(){return Style.Visible;};
    void setVisible(int visible){
      if (Style.Visible!=visible){
        Style.Visible=visible;
        updateFlags(WF_VISIBLE);
      }
    };

    void show(){setVisible(1);};
    void hide(){setVisible(0);};

    void redisplay(){
      bind();
      glutPostRedisplay();
    };

    virtual void onCreate(){};
    virtual void onDestroy(){};

    virtual void onDisplay(){
      glClearColor(randf(),randf(),randf(),0);
      glClear(GL_COLOR_BUFFER_BIT);
      glFlush();
      glutSwapBuffers();
    };

    virtual void onReshape(int width,int height){glViewport(0,0,width,height);};

    virtual void onShow(){};
    virtual void onHide(){};

    virtual void onMenuEnter(int x,int y){};
    virtual void onMenuLeave(int x,int y){};

    virtual void onIdle(){redisplay();};

    virtual void onKeyboard(unsigned char key,int x,int y){
      if (key==27)
        exit(0);
    };
    virtual void onSpecial(int key,int x,int y){};
#if (GLUT_API_VERSION >= 4 || GLUT_XLIB_IMPLEMENTATION >= 13)
    virtual void onKeyboardUp(unsigned char key,int x,int y){};
    virtual void onSpecialUp(int key,int x,int y){};
#endif

    virtual void onMouseDown(int button,int x,int y){};
    virtual void onMouseUp(int button,int x,int y){};
    virtual void onMouseEnter(){};
    virtual void onMouseLeave(){};
    virtual void onMouseMotion(int passive,int x,int y){}

    virtual void onButtonBoxDown(int button){};
    virtual void onButtonBoxUp(int button){};

    virtual void onDials(int dial,int value){};

    virtual void onSpaceballButtonDown(int button){};
    virtual void onSpaceballButtonUp(int button){};
    virtual void onSpaceballMotion(int x,int y,int z){};
    virtual void onSpaceballRotate(int x,int y,int z){};

    virtual void onTabletButtonDown(int button,int x,int y){};
    virtual void onTabletButtonUp(int button,int x,int y){};
    virtual void onTabletMotion(int x,int y){};
};

Window::Rect rect(int x,int y,int width,int height){
  Window::Rect R;
  R.X=x;
  R.Y=y;
  R.Width=width;
  R.Height=height;
  return R;
}

Window::WindowStyle windowStyle(int displayMode,unsigned int callBacks,int fullScreen=0,int visible=1){
  Window::WindowStyle W;
  W.DisplayMode=displayMode;
  W.CallBacks=callBacks;
  W.FullScreen=fullScreen;
  W.Visible=visible;
  return W;
}

const Window::Rect Window::DefTopWindowBounds={-1,-1,300,300};
const Window::Rect Window::DefSubWindowBounds={100,100,100,100};
const Window::WindowStyle Window::DefTopWindowStyle={
  GLUT_RGBA,
  WC_KEYBOARD | WC_MOUSE,
  1,
  0
};
const Window::WindowStyle Window::DefSubWindowStyle={
  GLUT_RGBA | GLUT_DOUBLE,
  WC_MOUSE,
  1,
  0
};

#endif

